#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "avrcore.h"

int flash_size = 8192;
uint16_t flash[8192];
int ram_size = 2048;
uint8_t ram[2048];

uSREG sreg;
uint16_t SP;
uint16_t PC;
uint8_t gpr[16];

uint8_t avr_core_mem_read (AvrCore *core, int addr) {
	if (addr >= 0x4000) {
		addr -= 0x4000;
		addr &= (flash_size - 1);
		if (addr & 1) return (flash[addr >> 1] >> 8) & 0xff;
		return flash[addr >> 1] & 0xff;
	}
	if (addr >= 0x40) {
		addr &= (ram_size - 1);
		return ram[addr];
	}
	return avr_core_io_read (core, addr);
}

void avr_core_mem_write (AvrCore *core, int addr, uint8_t val) {
	if (addr >= 0x4000) {
		return;
	}
	else if (addr >= 0x40) {
		addr &= (ram_size - 1);
		ram[addr] = val;
		printf("S mem[%04x] = %02x : PC %04x FLAGS %02x SP %04x\n"
			, addr , val, PC , sreg.reg, SP);

	}
	else
		avr_core_io_write (core, addr, val);
}

uint8_t avr_core_sreg_get (AvrCore *core) {
	return sreg.reg;
}
void avr_core_sreg_set (AvrCore *core, uint8_t v) {
	sreg.reg = v;
}

int avr_core_sreg_get_bit (AvrCore *core, int b) {
	return (sreg.reg & (1 << b) ) != 0;
}

void avr_core_sreg_set_bit (AvrCore *core, int b, int v) {
	sreg.reg = set_bit_in_byte(sreg.reg, b, v);
}

uint8_t avr_core_gpwr_get (AvrCore *core, int reg) {
	return gpr[reg & 0xf];
}
void avr_core_gpwr_set (AvrCore *core, int reg, uint8_t val, int as_idx) {
	gpr[reg & 0xf] = val;
	if (!as_idx)
		printf("L r%2d = %02x : PC %04x FLAGS %02x SP %04x\n"
			, reg , val, PC , sreg.reg, SP);
}

void avr_core_io_write (AvrCore *core, int addr, uint8_t val) {
	if (addr == 0x3f) sreg.reg = val;
	else if (addr == 0x3e) 
		SP = (SP & 0xff) | ((uint16_t)val << 8);
	else if (addr == 0x3d)
		SP = (SP & 0xff00) | (uint16_t)val;
}

uint8_t avr_core_io_read (AvrCore *core, int addr) {
	if (addr == 0x3f) return sreg.reg;
	if (addr == 0x3e) return (SP >> 8) & 0xff;
	if (addr == 0x3d) return SP  & 0xff;
	return 0;
}

uint32_t avr_core_stack_pop (AvrCore *core, int bytes) {
	uint32_t r = 0;
	while (bytes >= 1) {
		r <<= 8;
		++ SP;
		r |= avr_core_mem_read (core, SP);
		bytes --;
	}
	return r;
}

void avr_core_stack_push (AvrCore *core, int bytes, uint32_t val) {
	while (bytes >= 1) {
		avr_core_mem_write (core, SP, val & 0xff);
		val >>= 8;
		SP --;
		bytes --;
	}
}

int32_t avr_core_PC_size (AvrCore *core) {
	return 2;	// 2 bytes
}

int32_t avr_core_PC_get (AvrCore *core) {
	return PC;
}

void avr_core_PC_set (AvrCore *core, int32_t val) {
	PC = val;
}

void avr_core_PC_incr (AvrCore *core, int val) {
	PC = (PC + val) & 0xfff;
}

void avr_core_inst_CKS_set (AvrCore *core, int val) {
}

#  define strip_dir(path) (path)

void
private_avr_warning (char *file, int line, char *fmt, ...)
{
    va_list ap;
    char ffmt[128];

    snprintf (ffmt, sizeof (ffmt), "%s:%d: WARNING: %s", strip_dir (file),
              line, fmt);
    ffmt[127] = '\0';

    va_start (ap, fmt);
    vfprintf (stderr, ffmt, ap);
    va_end (ap);

    fflush (stderr);
}

void
private_avr_error (char *file, int line, char *fmt, ...)
{
    va_list ap;
    char ffmt[128];

    snprintf (ffmt, sizeof (ffmt), "\n%s:%d: ERROR: %s\n\n", strip_dir (file),
              line, fmt);
    ffmt[127] = '\0';

    va_start (ap, fmt);
    vfprintf (stderr, ffmt, ap);
    va_end (ap);

    fflush (stderr);

    exit (1);                   /* exit instead of abort */
}
