/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : USE_DMY_CLOCK, PRC_WATCH, CHECK_CONFLICT

//`define USE_DMY_CLOCK
//`define PRC_WATCH
`define FOWARD_DI
//`define CHECK_CONFLICT
//`define GPR_INITCLR

// GPR modele for the Reduced Core TinyAVR (attiny10) 
// version 005
// 
// Target Device: xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal: Balanced
// LATCH_DOL : Insert LATCHes for Timing
//                                      NORMAL    GPR_WB_TUNE
// Number of Slice Flip Flops:              51          51
// Number of 4 input LUTs:                  92         112
// Number of occupied Slices:               66          71
// Total Number of 4 input LUTs:           100         113
// Number used for Dual Port RAMs:          16          16
// Number of bonded IOBs:                   90          90
// Post-PAR Static Timing Report
// Clock to Setup (CLK2X) MAX           10.241       7.468              
//
// NORMAL : GPR_2X_ONLY (for unit_test)
//
module rtavr_gpr_16(
		input RESET,
`ifdef USE_DMY_CLOCK
		input CLK2X_IN,
		output CLK_OUT,
`else
		input CLK2X,
		input CLK,
`endif
		input WE,
		input PREDEC, POSTINC, 
		input [7:0] DI,
		input [3:0] ADDRAL,
		input [3:0] ADDRAH,
		input [3:0] ADDRBL,
		input [3:0] ADDRBH,
		output [7:0] DOBL,
		output [7:0] DOBH
`ifdef GPR_2X_ONLY
		, input [3:0] r_addrbl
		, input [3:0] r_addrbh
		, input [7:0] s2_DOBL
`endif
`ifdef CHECK_CONFLICT
		, output DI_CONFLICT
`endif
		, output [11:0] JA
		, output [15:0] INDEX
`ifdef PRC_WATCH
		, output [3:0] WB_ADDR
`endif
		);
// Timing MEMO:
//
//                   _______________             ______________
// CLK2X ____________|              |____________|             |
//                                  ___________________________
// CLK  ____________________________|                          |
//
//
//      |            ------  ALL input --------                 |
//    ---- old value  ---           |  JA/INDEX low byte        |
//                                  |---|  JA       high byte   |
//                                  |----------|INDEX high byte |
//      |                      DOBL(old value)                  | DOBL
//      |                      DOBH(old value)                  | DOBH

`ifdef USE_DMY_CLOCK
  reg CLK;
  wire CLK2X;
  assign CLK_OUT = CLK;
  assign CLK2X   = CLK2X_IN;
`endif
  reg  r_clk;
  reg  [3:0] r_addra;
  reg  [3:0] r_addrb;
`ifdef GPR_2X_ONLY
`else
  reg  [3:0] r_addrbl;
  reg  [3:0] r_addrbh;
  reg  [7:0] s2_DOBL;
`endif
  reg  [7:0] s2_DOBH;
  reg  [7:0] gpr [0:15];

  assign DOBL = s2_DOBL;
  assign DOBH = s2_DOBH;

  wire [8:0] DOA = { 1'b0, gpr[r_addra] };

  reg  c;

`ifdef CHECK_CONFLICT
  assign DI_CONFLICT = WE & (ADDRAL == ADDRAH);
  wire WB = (PREDEC | POSTINC) & ~DI_CONFLICT;
`else
  wire WB = (PREDEC | POSTINC);
`endif
  reg  [3:0] r_wb_addr;
  `ifdef PRC_WATCH
    assign WB_ADDR = r_wb_addr;
  `endif

 
`ifdef GPR_WB_TUNE
  wire [7:0] DOA2 = (PREDEC & (~CLK | c) ) ? DOA - 1 : DOA;
  wire [8:0] WB_DATA = (PREDEC & (~CLK | c) ) ?  {1'b0 , DOA} - 1
                     : (POSTINC & (~CLK | c)) ?  {1'b0 , DOA} + 1
		     : {1'b0 , DOA};
`else
  wire [8:0] DOA2 = {1'b0 , DOA} - (PREDEC & (~CLK | c));
  wire [8:0] WB_DATA = DOA2 + (POSTINC & (~CLK | c));
`endif
  
  wire WE2 = WB | (WE & CLK) ;
  wire [7:0] DI2 = (WB) ? WB_DATA : DI ;

  reg  [7:0] s1_DOAL;
  reg  [7:0] s1_DOAH;
  reg  [7:0] s1_DOBL;

`ifdef FOWARD_DI
  wire FORWARD_TO_BL = (WE) & (r_wb_addr == r_addrbl);
  wire FORWARD_TO_BH = (WE) & (r_wb_addr == r_addrbh);

  wire [7:0] v_DOBL = FORWARD_TO_BL ? DI2 : s1_DOBL;
  wire [7:0] v_DOBH = FORWARD_TO_BH ? DI2 : gpr[r_addrb];
`else
  wire [7:0] v_DOBL = s1_DOBL;
  wire [7:0] v_DOBH = gpr[r_addrb];
`endif

  assign JA[7:0] = v_DOBL;
  assign JA[11:8] = v_DOBH[3:0];
  assign INDEX[7:0] = s1_DOAL;
`ifdef NO_STALL_IDX
  assign INDEX[15:8] = (WB) ? s1_DOAH : v_DOBH ;
`else
  assign INDEX[15:8] = s1_DOAH;
`endif

`ifdef USE_DMY_CLOCK
always @(negedge CLK2X)
  begin
	CLK <= ~CLK;
  end

initial
  begin
	CLK <= 0;
  end
`endif

always @(posedge CLK2X)
  begin
    if (RESET == 0)
      begin
	r_clk <= 0;
	c <= 0;
	s1_DOAL <= 0;
	s1_DOAH <= 0;
	r_wb_addr <= 0;
      end
    else
      begin
	r_clk <= CLK;
	if (WE2) gpr[r_addra] <= DI2;
	if (~CLK) s1_DOAL <= DOA2[7:0];
	if (CLK) s1_DOAH <= DOA2[7:0];
`ifdef NO_STALL_IDX
	if (~CLK) r_wb_addr <= (WE) ? ADDRAH : ADDRBH;
`else
	if (~CLK) r_wb_addr <= (WE) ? ADDRAH : ADDRAL + 1;
`endif
	c <= WB_DATA[8];
      end
  end

always @(negedge CLK2X)
  begin
    if (RESET == 0)
      begin
	r_addra <= 0;
	r_addrb <= 0;
	s1_DOBL <= 0;
	s2_DOBH <= 0;
      end
    else
      begin
	r_addra <= r_clk ? ADDRAL : r_wb_addr;
	r_addrb <= r_clk ? ADDRBL : ADDRBH;
	if (!r_clk) s1_DOBL <= gpr[r_addrb];
	if (r_clk) s2_DOBH <= v_DOBH;
      end
  end

`ifdef GPR_2X_ONLY
`else
  always @(negedge CLK)
  begin
    if (RESET == 0)
      begin
	s2_DOBL <= 0;

	r_addrbl <= 0;
	r_addrbh <= 0;
      end
    else
      begin
	s2_DOBL <= v_DOBL;

	r_addrbl <= ADDRBL;
	r_addrbh <= ADDRBH;
      end
  end
`endif

`ifdef GPR_INITCLR
  integer i;
  initial
    begin
      for (i = 0 ; i < 16 ; i = i + 1)
	gpr[i] <= 8'h00;
    end
`endif
endmodule
